"""
oneline.core is the set of main functions and features, which contains:

    1. OneData, extended DataFrame
    2. OneSeries, extended Series
    3. plot(private), a plot function for fast using
"""

from .onedata import OneData
from .oneseries import OneSeries



"""
The compatibility functions for some of the oneline modules.
"""

import sys
import platform
import importlib


def import_optional_dependency(name: str, error_raise: bool = True):
    """
    Import function and raise ImportError if not exists.

    :param name: the name of module
    :param error_raise: raise the import error
    :return: module
    """

    try:
        module = importlib.import_module(name)
    except ImportError:
        if error_raise:
            raise ImportError(f"Missing optional dependency '{ name }'") from None
        else:
            return None

    return module


def is_platform_little_endian():
    """
    Checking if the running platform is little endian.

    :return: bool
    """

    return sys.byteorder == "little"


def is_platform_windows():
    """
    Checking if the running platform is windows.

    :return: bool
    """

    return sys.platform in ["win32", "cygwin"]


def is_platform_linux():
    """
    Checking if the running platform is linux.

    :return: bool
    """

    return sys.platform == "linux"


def is_platform_mac():
    """
    Checking if the running platform is mac.

    :return: bool
    """

    return sys.platform == "darwin"


def is_platform_arm():
    """
    Checking if he running platform use ARM architecture.

    :return: bool
    """

    return platform.machine() in ("arm64", "aarch64")



"""
A set of methods that can make conversion related codes easier, faster and safer. This module is based on
built-in md5 and base64, and some of the methods can prevent memory overflow.
"""

from hashlib import md5
import base64


def file2md5(path: str, batch_byte=4096):
    """
    A md5 hash function that have the ability to prevented memory overflow.

    :param path: the path of file
    :param batch_byte: the size of every batch
    :return: a string of md5
    """

    md5_obj = md5()
    with open(path, 'rb') as f:
        while True:
            data = f.read(batch_byte)
            if not data:
                break
            md5_obj.update(data)

    return md5_obj.hexdigest()


def str2md5(s: str):
    """
    A sister function for fast using.

    :param s: the string which will be hashed
    :return: a string of md5
    """

    return md5(s.encode()).hexdigest()


def file2base64(path: str, batch_byte=4096):
    """
    A base64 hash function that have the ability to prevented memory overflow.

    :param path: the path of file
    :param batch_byte: the size of every batch
    :return: base64 bytes result
    """

    base64_list = bytes()
    with open(path, 'rb') as f:
        while True:
            data = f.read(batch_byte)
            if not data:
                break
            base64_list += base64.b64encode(data)

    return base64_list


def str2base64(s: str):
    """
    A sister function for fast using.

    :param s: the string which will be converted to base64
    :return: a string of based64
    """

    return base64.b64encode(s.encode())


def base642file(s: str, path: str):
    """
    Generate a file from base64 string.

    :param s: the base64 string
    :param path: the path of file
    :return: None
    """

    with open(path, "wb")as f:
        f.write(base64.b64decode(s))


def base642str(s: str):
    """
    Convert a base64 to string.

    :param s: the base64 string
    :return: a string
    """

    return base64.b64decode(s).decode()



"""
It's a personal code set for learning and researching, and it provides advanced encapsulation of some
libraries, designing to provide a simpler mode of operation and help people to simplify their codes
concisely.

WARNING: Since this library is a collection of code used for learning and analysis, the content is not
perfect for the time being.
"""

# Core
from .core import *

# MachineLearning
from .machinelearning import *

# Information
__version__ = '0.2.7'
__author__ = 'Zeesain Tsui'

# check the essential dependencies
hard_dependencies = ("numpy", "pandas", "matplotlib")
missing_dependencies = []

for dependency in hard_dependencies:
    try:
        __import__(dependency)
    except ImportError as e:
        missing_dependencies.append(f"{dependency}: {e}")

if missing_dependencies:
    raise ImportError(
        "Unable to import required dependencies:\n" + "\n".join(missing_dependencies)
    )
del hard_dependencies, dependency, missing_dependencies



"""
An error class for raising different errors with clear information.
"""


class Error(object):

    def __init__(self):
        return

    def _input_error(self):
        """
        The exception of ValueError when format was unsupported.

        :return: ValueError
        """

        raise ValueError('Input error, please input a valid data that satisfied.')

    def _raise_parameter_error(self, args):
        """
        The exception of ValueError when format was unsupported.

        :return: ValueError
        """

        raise ValueError('Input parameter {} {} not exist.'.format(", ".join(args), "does" if len(args) == 1 else "do"))

    def _wrong_parameter_error(self, args):
        """
        The exception of ValueError when the input parameters were wrong.

        :return: ValueError
        """

        raise ValueError('The input parameter {} {} the wrong value.'.format(", ".join(args), "has" if len(args) == 1 else "have"))

    def _raise_plot_value_error(self, s: list):
        """
        A ValueError would be raised if required value was missed.

        :param s: the missing value(s)
        :return: None
        """

        raise ValueError(f'The parameter { ", ".join(s) } {"is" if len(s) == 1 else "are"} required.')

    def _raise_plot_format_error(self, s: list, format: str):
        """
        A ValueError of format would be raised if the format was not matched.

        :param s: the wrong parameters
        :param format: the required format
        :return: None
        """

        raise ValueError(f'The parameter { ", ".join(s) } should be { format}, rather than { ", ".join([str(type(n)) for n in s]) }.')



# import pd as reference
import pandas as pd

# import OneData
from oneline import OneData

# input the dataset using simple OneData() with file location
data = OneData('./gender_submission.csv')

# rebuilt print function
print(data)

# OneData is compatible with DataFrame
data_from_pd = pd.read_csv('gender_submission.csv')
data = OneData(data_from_pd)

# remove columns or rows using remove method
data = data.remove(column=['PassengerId'])

# reduce the memory usage of data
data.reduce_mem_usage(info=True)
# summarize the data
data.summary()

# plot function
data.count_plot("Gender")

# The function of DataFrame is also valid in OneData
data.to_csv('test_new.csv', index=False)
data = data.drop('PassengerId')  # drop() from DataFrame



"""
oneline.machinelearning is the set of functions and features for machine learning and extended deep
learning, which contains:

    1. AverageMeter, a average meter for counting and analyze the data during training
    2. NeuralNetwork, extended deep learning module for fast training
    3. MachineLearning, a set of useful classical machine learning methods
    4. MachineLearningModel, a container for machine learning training
"""

from .average_meter import AverageMeter
from .neural_network import NeuralNetwork
from .machine_learning import *



"""
The IO module supplies some basic IO methods with a more comprehensive auto detection. It's easy to use
without learning cost.

This module are still in progress.
"""

from .io import *
from .file_input import *



"""
A wrapper for classical machine learning based on sci-kit learn. Because this simple wrapper is not
necessary, so this module would be deleted in the future.
"""

from ..tools.compat import import_optional_dependency


class MachineLearningModel(object):
    def __init__(self, model):
        self.model = model

    def fit_model(self, x, y, info):
        self.model.fit(x, y)
        self.model.score(x, y)
        if info:
            print('Coefficient: \n', self.model.coef_)
            print('Intercept: \n', self.model.intercept_)

    def predict(self, x_test: list):
        """
        Predict method for prediction.

        :param x_test: the data for prediction
        """

        return self.model.predict(x_test)


def linear_regression(x, y, degree: int = 1, include_bias: bool = False, info: bool = False):
    """
    Linear Regression Method.

    :param x: the x
    :param y: the y
    :param degree: the degree
    :param include_bias: include bias or not
    :param info: show the info and parameters of the model
    :return: model
    """

    # set the degree
    if degree > 1:
        pf = import_optional_dependency("sklearn.preprocessing")\
            .PolynomialFeatures(degree=degree, include_bias=include_bias)
        x = pf.fit_transform(x)

    # fit the model
    model = MachineLearningModel(import_optional_dependency("sklearn.linear_model").LinearRegression())
    model.fit_model(x, y, info)
    return model


def logistic_regression(x, y, info: bool = False):
    """
    Logistic Regression Method.

    :param x: the x
    :param y: the y
    :param info: show the info and parameters of the model
    :return: model
    """

    # fit the model
    model = MachineLearningModel(import_optional_dependency("sklearn.linear_model").LogisticRegression())
    model.fit_model(x, y, info)
    return model


def decision_tree(x, y, criterion: str = "gini"):
    """
    Decision Tree Method.

    :param x: the x
    :param y: the y
    :param criterion: the criterion method
    :return: model
    """

    # fit the model
    model = MachineLearningModel(
        import_optional_dependency("sklearn.tree").DecisionTreeClassifier(criterion=criterion))
    model.fit_model(x, y, info=False)
    return model


def svm(x, y, C: float = 1.0, kernel: str = 'rbf'):
    """
    Support Vector Machine Method.

    :param x: the x
    :param y: the y
    :param C: the Regularization parameter
    :param kernel: Specifies the kernel type to be used in the algorithm
    :return: model
    """

    # fit the model
    model = MachineLearningModel(import_optional_dependency("sklearn.svm").SVC(C=C, kernel=kernel))
    model.fit_model(x, y, info=False)
    return model


def naive_bayes(x, y):
    """
    Naive Bayes Method.

    :param x: the x
    :param y: the y
    :return: model
    """

    # fit the model
    model = MachineLearningModel(import_optional_dependency("sklearn.naive_bayes").GaussianNB())
    model.model.fit(x, y)
    return model


def knn(x, y, n_neighbors=6):
    """
    KNN methods.

    :param x: the x
    :param y: the y
    :param n_neighbors: the number of n-neighbors
    :return: model
    """

    # fit the model
    model = MachineLearningModel(
        import_optional_dependency("sklearn.neighbors").KNeighborsClassifier(n_neighbors=n_neighbors))
    model.model.fit(x, y)
    return model


def kmeans(x, n_clusters=6, random_state=0):
    """
    K-Means Method.

    :param x: the x
    :param n_clusters: the number of clusters
    :param random_state: the number of random state
    :return: model
    """

    # fit the model
    model = MachineLearningModel(
        import_optional_dependency("sklearn.cluster").KMeans(n_clusters=n_clusters, random_state=random_state))
    model.model.fit(x)
    return model


def random_forest(x, y):
    """
    Random Forest Method.

    :param x: the x
    :param y: the y
    :return: model
    """

    model = MachineLearningModel(import_optional_dependency("sklearn.ensemble").RandomForestClassifier())
    model.model.fit(x, y)
    return model



"""
A super lightweight module for Neural Network training, which now only includes basic training methods and
will be updated in the future.
"""

from os.path import join, isdir, isfile

from .average_meter import AverageMeter
from ..tools.compat import import_optional_dependency


class NeuralNetwork(object):
    """
    A tools for a fast training process of neural network, which can start training without other code after
    the dataset, optimizer and other parameters declared.
    """

    def _clean_cache(self):
        """
        Clean the cache before auto-training, which can avoid .

        :return: None
        """

        torch = import_optional_dependency("torch")
        if self.device == torch.device('cuda'):
            with torch.cuda.device(self.device):
                torch.cuda.empty_cache()

    def _raise_format_error(self, name: str, format_str: str, source_format: str):
        """
        The exception of ValueError when format was unsupported.

        :return: ValueError
        """

        raise ValueError(f"The '{ name }' should be { format_str }, rather than { source_format }")

    def __init__(self,
                 model=None,
                 train_dataset=None,
                 eval_dataset=None,
                 optimizer=None,
                 criterion=None,
                 cpu: bool = False):
        """
        Initialize the Neural Network training process.

        :param model: the torch.nn.Module
        :param train_dataset: the DataLoader() dataset for training
        :param eval_dataset: the DataLoader() dataset for testing
        :param optimizer: the torch optimizer
        :param criterion: the torch criterion
        :param cpu: set True if forced to use CPU, else it would be set automatically
        """

        # import torch for initialization
        torch = import_optional_dependency("torch")

        # ============== basic parameters ============== #
        # the device that used to train models, which can automatically set
        self.device = torch.device("cuda" if torch.cuda.is_available() and not cpu else "cpu")
        # the optimizer of training
        self.optimizer = optimizer
        # the neural network model
        self.model = model.to(self.device) if model else None
        # the criterion of training
        self.criterion = criterion.to(self.device) if criterion else None
        # the dataset for training
        self.train_dataset = train_dataset
        # the dataset for evaluation
        self.eval_dataset = eval_dataset
        # the training process would show information if self.info is True
        self.info = True

        # ============== the parameters of training ============== #
        # the loss average meter for every epoch
        self.epoch_loss = AverageMeter()
        # the counter for training
        self.epoch = 0
        # training process for iteration
        self.batch_process = None

    def _set_train(self):
        """
        A sub-function that ensuring the train mode.

        :return: None
        """

        if not self.model.__dict__['training']:
            self.model.train()

    def _set_eval(self):
        """
        A sub-function that ensuring the eval mode.

        :return: None
        """

        if self.model.__dict__['training']:
            self.model.eval()

    @staticmethod
    def _set_save_location(location):
        if isdir(location):
            return join(location, 'model.pkl')
        return location

    def _batch_iter(self, source, target, i: int):
        """
        The train function that executes a standard training flow per epoch.

        :return:
        """
        # send data to device
        source = source.to(self.device)
        target = target.to(self.device)

        # the result and loss
        result = self.model(source)
        loss = self.criterion(result, target)

        # optimization and backward
        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

        # update the loss
        self.epoch_loss.update(loss.item(), source.size(0))

        # print the information
        if self.info:
            print(f"\rEpoch: { self.epoch } | Batch: { i } | loss: { self.epoch_loss.avg }", end="")

        # clean the data
        del source, target

        return result

    def _train_batch(self):
        """
        The train iterator that executes a standard training flow per batch.

        :return:
        """

        # start epoch
        for i, (source, target) in enumerate(self.train_dataset):
            result = self._batch_iter(source, target, i)

            # yield
            yield result

    @staticmethod
    def _reset_weights(m):
        """
        A sub-function with a general weights initialization.

        :param m: the layers of self.model
        :return: None
        """

        nn = import_optional_dependency("torch.nn")
        init = import_optional_dependency("torch.nn.init")
        if isinstance(m, nn.Conv1d):
            init.normal_(m.weight.data)
            if m.bias is not None:
                init.normal_(m.bias.data)
        elif isinstance(m, nn.Conv2d):
            init.xavier_normal_(m.weight.data)
            if m.bias is not None:
                init.normal_(m.bias.data)
        elif isinstance(m, nn.Conv3d):
            init.xavier_normal_(m.weight.data)
            if m.bias is not None:
                init.normal_(m.bias.data)
        elif isinstance(m, nn.ConvTranspose1d):
            init.normal_(m.weight.data)
            if m.bias is not None:
                init.normal_(m.bias.data)
        elif isinstance(m, nn.ConvTranspose2d):
            init.xavier_normal_(m.weight.data)
            if m.bias is not None:
                init.normal_(m.bias.data)
        elif isinstance(m, nn.ConvTranspose3d):
            init.xavier_normal_(m.weight.data)
            if m.bias is not None:
                init.normal_(m.bias.data)
        elif isinstance(m, nn.BatchNorm1d):
            init.normal_(m.weight.data, mean=1, std=0.02)
            init.constant_(m.bias.data, 0)
        elif isinstance(m, nn.BatchNorm2d):
            init.normal_(m.weight.data, mean=1, std=0.02)
            init.constant_(m.bias.data, 0)
        elif isinstance(m, nn.BatchNorm3d):
            init.normal_(m.weight.data, mean=1, std=0.02)
            init.constant_(m.bias.data, 0)
        elif isinstance(m, nn.Linear):
            init.xavier_normal_(m.weight.data)
            init.normal_(m.bias.data)
        elif isinstance(m, nn.LSTM):
            for param in m.parameters():
                if len(param.shape) >= 2:
                    init.orthogonal_(param.data)
                else:
                    init.normal_(param.data)
        elif isinstance(m, nn.LSTMCell):
            for param in m.parameters():
                if len(param.shape) >= 2:
                    init.orthogonal_(param.data)
                else:
                    init.normal_(param.data)
        elif isinstance(m, nn.GRU):
            for param in m.parameters():
                if len(param.shape) >= 2:
                    init.orthogonal_(param.data)
                else:
                    init.normal_(param.data)
        elif isinstance(m, nn.GRUCell):
            for param in m.parameters():
                if len(param.shape) >= 2:
                    init.orthogonal_(param.data)
                else:
                    init.normal_(param.data)

    def __repr__(self):
        """
        Show the information of training.

        :return: None
        """

        # info string
        info = self.model.__repr__()
        info += "\n=========================\n"
        info += f"Train data length:\t\t{ len(self.train_dataset) }\n"
        info += f"Eval sata length:\t\t{ len(self.eval_dataset) }\n"
        info += f"Optimizer:\t\t\t\t{ str(self.optimizer).split('(')[0] }\n"
        info += f"Criterion:\t\t\t\t{ str(self.criterion).split('(')[0] }\n"
        info += f"Training Environment:\t{ self.device.type }\n"
        info += f"Show information:\t\t{ 'True' if self.info else 'False' }\n"
        info += "=========================\n"

        return info

    @property
    def layers(self):
        return self.model.state_dict().items()

    def auto_train(self,
                   epoch: int,
                   save_model_location: str,
                   eval: bool = True,
                   eval_interval: int = 1,
                   info: bool = True,
                   save_static_dicts: bool = True):
        """
        A auto training method for fast using. And the epoch and the location for saving models should be
        specified while using auto_train().

        :param epoch: teh epoch of training
        :param save_model_location: the location for storing the model
        :param eval: eval the model or not
        :param eval_interval: the interval epoch between each eval processes
        :param info: Show the information or not
        :param save_static_dicts: set True if only save the static dicts
        :return: None
        """

        # initialization
        self.info = info
        best_attempt = float("-inf")
        self._set_train()

        # clean the cache if cuda is available
        self._clean_cache()

        # start training
        for n in range(epoch):
            self.iter_epoch()
            if self.eval_dataset and eval and not (n + 1) % eval_interval:
                # eval start
                eval_loss, accuracy = self.eval()

                # save the best model
                if accuracy > best_attempt:
                    best_attempt = accuracy
                    if save_static_dicts:
                        self.save_state_dict(save_model_location)
                    else:
                        self.save_model(save_model_location)
                    print(f"The best model is saved to { self._set_save_location(save_model_location) }. "
                          f"Best accuracy: { best_attempt }")

    def iter_epoch(self):
        """
        An function that would stop training per epoch for advanced calculation, plot, etc.

        :return: None
        """

        # set to train mode
        self._set_train()

        # start epoch
        for i, (source, target) in enumerate(self.train_dataset):
            self._batch_iter(source, target, i)

        if self.info:
            print(f"\rEpoch: { self.epoch } | Average loss: { self.epoch_loss.avg }")

        # update epoch and reset the epoch_loss
        self.epoch_loss.reset()
        self.epoch += 1

    def iter_batch(self):
        """
        An iterator that training per batch and return for advanced calculation, plot, etc.

        :return: output
        """

        # model initialization
        self._set_train()

        if not self.batch_process:
            self.batch_process = self._train_batch()
            return self.batch_process.__next__()
        else:
            try:
                return self.batch_process.__next__()
            except StopIteration:
                # update the state if StopIteration
                if self.info:
                    print(f"\rEpoch: { self.epoch } | Average loss: { self.epoch_loss.avg }")

                # update epoch and reset the epoch_loss
                self.epoch_loss.reset()
                self.epoch += 1

                # reset the batch process
                del self.batch_process
                self.batch_process = self._train_batch()
                return self.batch_process.__next__()

    def eval(self):
        """
        The evaluation flow using test_dataset without grad.

        :return: The total loss during evaluation and model's accuracy
        """

        # parameters initialize
        torch = import_optional_dependency("torch")
        eval_total = 0
        eval_correct = 0
        eval_loss = 0
        self._set_eval()

        # display the information
        if self.info:
            print(f"\rEvaluating...", end="")

        # start eval part
        for i, (source, target) in enumerate(self.eval_dataset):
            # send data to device
            source = source.to(self.device)
            target = target.to(self.device)

            result = self.model(source)
            eval_loss += self.criterion(result, target).item()
            _, predicted = torch.max(result.data, 1)
            eval_total += target.size(0)
            eval_correct += (predicted == target).sum().item()

        accuracy = eval_correct / eval_total
        eval_loss = eval_loss / eval_total

        if self.info:
            print(f"\rEvaluation loss: { eval_loss } | Accuracy: { accuracy }")

        return eval_loss, accuracy

    def reset_train(self):
        """
        Reset the process of training, which includes the loss meter reset, epoch reset and model's weights
        reset.

        :return: None
        """

        self.model.apply(self._reset_weights)
        self.epoch_loss.reset()
        self.epoch = 0
        del self.batch_process
        self.batch_process = None

    def save_weights(self, location: str):
        """
        Save only the state dict of the model.

        :param location: the location of models
        :return: None
        """

        # import torch
        torch = import_optional_dependency("torch")
        torch.save(self.model.state_dict(), self._set_save_location(location))

    def save_model(self, location: str):
        """
        Save only the whole model.

        :param location: the location of models
        :return: None
        """

        # import torch
        torch = import_optional_dependency("torch")

        torch.save(self.model, self._set_save_location(location))

    def load_weights(self, location: str):
        torch = import_optional_dependency("torch")
        if not isfile(location):
            raise ValueError(f"The { location } is not a valid dict file.")
        self.model.load_state_dict(torch.load(location))

    def reset_weights(self):
        self.model.apply(self._reset_weights)

    def check_parameters(self):
        """
        A method for checking the parameters before training, in order to process the training correctly.

        :return: None
        """

        torch = import_optional_dependency('torch')
        if not isinstance(self.model, torch.nn.Module):
            self._raise_format_error('self.model', 'torch.nn.Module', f'{ type(self.model) }')
        if not isinstance(self.optimizer, torch.optim.Optimizer):
            self._raise_format_error('self.optimizer', 'torch.optim.Optimizer', f'{ type(self.optimizer) }')
        if not isinstance(self.train_dataset, torch.utils.data.DataLoader):
            self._raise_format_error('self.train_dataset', 'torch.utils.data.DataLoader', f'{ type(self.train_dataset) }')
        if not isinstance(self.eval_dataset, torch.utils.data.DataLoader):
            self._raise_format_error('self.eval_dataset', 'torch.utils.data.DataLoader', f'{ type(self.eval_dataset) }')



"""
The plot function of oneline. It contains a series of plot methods for a fast plot using.
"""

import numpy as np
import matplotlib.pyplot as plt
from ..tools.compat import import_optional_dependency


class Plot(object):
    """
    This is a plot module for data visualization.
    """

    @property
    def _plt(self):
        import matplotlib.pyplot as plt
        return plt

    def _plot_prev_config(self, inherit: plt = None,
                          figsize: list = None,
                          style: str = None):
        """
        A general pre-configuration of plot process.

        :param inherit: the outscope configuration of plot
        :param figsize: the size of figuration
        :param style: the style of plot
        :return: plt
        """

        # inherit previous plt configuration if exists
        if inherit:
            plt = inherit
        else:
            plt = self._plt
            plt.style.use(style)

        # set the size of figuration if it's required
        if figsize:
            plt.figure(figsize=figsize)

        return plt

    @staticmethod
    def _plot_post_config(plt,
                          legend_loc: str,
                          title: str,
                          xlabel: str,
                          ylabel: str,
                          show: bool):
        """
        The post-configuration of plot process.

        :param plt: the plt
        :param legend_loc: the position of legend
        :param title: the title
        :param xlabel: the xlabel
        :param ylabel: the ylabel
        :param show: show the plot if True
        :return: plt
        """

        # Other post configuration
        if legend_loc:
            plt.legend(loc=legend_loc)
        if title:
            plt.title(title)
        if xlabel:
            plt.xlabel(xlabel)
        if ylabel:
            plt.ylabel(ylabel)

        # show the plot if turns True
        if show:
            plt.show()

        # return for advanced adjustment
        return plt

    @staticmethod
    def _meta_line_plot(plt, y_val, x, smooth, kind, interval, name):
        """
        A meta function of line_plot.

        :param plt: the inherit plt
        :param y_val: y value
        :param x: x value
        :param smooth: smooth the line if True
        :param kind: the kind of smooth method
        :param interval: the number of interval
        :param name: the name of line
        :return: plt
        """

        if smooth:
            interpolate = import_optional_dependency("scipy.interpolate")
            x_new = np.linspace(min(x), max(x), len(x) * interval)
            y_smooth = interpolate.interp1d(x, y_val, kind=kind)
            plt.plot(x_new, y_smooth(x_new), label=name)
        else:
            plt.plot(x, y_val, label=name)

        return plt



"""
The load_file module supplies some basic IO methods with a more comprehensive auto detection. It's easy to use
without learning cost.

This module are based on the Python standard library shutil.
"""

import json


def load_json(path: str, encoding='utf-8'):
    """
    Load the json file, which will return a dictionary.
    """

    with open(path, 'r', encoding=encoding) as f:
        return json.load(f)


def save_json(_dict: dict, path: str, indent=4, ensure_ascii=False, encoding='utf-8'):
    """
    Save a dictionary to json file
    """

    with open(path, 'w', encoding=encoding) as f:
        f.write(json.dumps(_dict, indent=indent, ensure_ascii=ensure_ascii))



"""
A set of methods that can make conversion related codes easier, faster and safer. This module is based on
built-in md5 and base64, and some of the methods can prevent memory overflow.

This module are still in progress.
"""

from .hash import *




def isPrime(number):
    if number % 2 == 0:
        return False
    for i in range(3, int(number ** 0.5) + 1, 2):
        if number % i == 0:
            return False
    return True

def isOdd(number):
    return number % 2

def isEven(number):
    return not number % 2



"""
The io module supplies some basic IO methods with a more comprehensive auto detection. It's easy to use
without learning cost.

This module are based on the Python standard library shutil.
"""

import os
import shutil
from os.path import isfile, isdir, exists


def ignore(*ignorance):
    """
    A built-in ignore_patterns function.
    """

    return shutil.ignore_patterns(*ignorance)


def copy(source: str, to: str, type=None, ignore=None, follow_symlinks=True):
    """
    Copy a file or a folder to destination, which can be a folder or a file (automatically detect if type
    isn't specified), and you can use io.ignore to ignore the particular files from copy process.
    ** For preventing conflicts, type can be specified when needed.

    For example: io.copy('input', 'new', ignore=io.ignore('*.csv'))
    """
    if type == "file":
        shutil.copy(source, to, follow_symlinks=follow_symlinks)
    elif type == "folder" or isdir(source):
        if ignore:
            shutil.copytree(source, to, ignore=ignore)
        else:
            shutil.copytree(source, to)
    elif isfile(source):
        shutil.copy(source, to, follow_symlinks=follow_symlinks)
    else:
        raise TypeError(f"The file or directory *{ source }* is not exist.")


def delete(source):
    """
    Delete the file or folder.
    """

    shutil.rmtree(source)


def move(source: str = None, to: str = None):
    """
    Move file or folder.
    """

    shutil.move(source, to)


def mkdir(path: str):
    """
    Create a new directory.
    """

    if exists(path):
        raise AttributeError(f"The path *{ path }* is exist.")
    else:
        os.mkdir(path)


def read(path: str, mode: str = 'r', encoding: str = 'utf-8', errors=None):
    """
    That's a simple and fast read function for reading a file.

    WARNING: This is for only small size of files, which means that the bigger size will lead to a memory
    exception.
    """

    with open(path, mode, encoding=encoding, errors=errors) as f:
        return f.read()


def readlines(path: str, mode: str = 'r', encoding: str = 'utf-8', errors=None, ):
    """
    That's a simple and fast readlines function for reading a file.

    WARNING: This is for only small size of files, which means that the bigger size will lead to a memory
    exception.
    """

    with open(path, mode, encoding=encoding, errors=errors) as f:
        return f.readlines()


def write(content, path, mode: str = 'w+', encoding: str = 'utf-8', errors=None):
    """
    Write string or list to file.

    If the content is a list, it will be written with writelines mode.
    And also the write mode will be used if the content is string.
    """

    with open(path, mode, encoding=encoding, errors=errors) as f:
        if isinstance(content, list):
            f.writelines(content)
        elif isinstance(content, str):
            f.write(content)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='one-line',
    version='0.2.7',
    description='Make every step oneLine.',
    long_description=long_description,  # 没有 README 这一项可以不要
    long_description_content_type="text/markdown",
    install_requires=[  # 该包的依赖包，安装时会检查是否满足依赖要求
        'pandas'
    ],
    packages=setuptools.find_packages(),  # 该库中含有的包，一般就自动搜索
    author='Zeesain Tsui',
    author_email='clarenceehsu@163.com',
    url='https://github.com/clarenceehsu/oneLine',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ],
)


class AverageMeter(object):
    """
    An average meter for counting and recording the data during process.
    """

    def __init__(self):
        """
        Initialize the meter.
        """

        # the average of the recording parameter
        self.avg = 0
        # the overall sum of the recording parameter
        self.sum = 0
        # the counter of the recording parameter
        self.count = 0

    def reset(self):
        """
        Reset the average meter to the initial.

        :return: None
        """

        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        """
        Update the record.

        :param val: the value of parameter
        :param n: number of the value that should be counted
        :return: None
        """

        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count



"""
OneData is an extended DataFrame based on pandas.DataFrame, which also inherits all the pandas.DataFrame
features and ready to use. It contains many useful methods for a better experience on data analysis.

WARNING: Because this module is still pre-alpha, so many features are unstable.
"""

import collections
import numpy as np
from abc import ABC
import matplotlib.pyplot as plt
import matplotlib.ticker as mtick

# pandas
import pandas as pd
from pandas import DataFrame
from pandas._libs import lib
from pandas.api.types import is_categorical_dtype
from pandas.api.types import is_datetime64_any_dtype as is_datetime

# pandas.core
from pandas.core import common as com
from pandas.core.generic import NDFrame
from pandas.core.dtypes.common import (
    is_hashable,
    is_integer,
    is_iterator,
    is_list_like,
)
from pandas.core.dtypes.generic import ABCMultiIndex
from pandas.core.indexing import convert_to_index_sliceable

# oneline module
from .plot import Plot
from .error import Error
from .oneseries import OneSeries
from ..tools.compat import import_optional_dependency


class OneData(DataFrame, ABC, Plot, Error):
    """
    That's the initial of the OneData class.
    Acceptable type:
        1. OneData
        2. DataFrame
        3. List
        4. Numpy Array
        5. String (the string should be a address of particular file, which can be automatically input based on format
                   by OneData)
    OneData is a advanced class of Pandas.DataFrame with more methods and all the features of DataFrame inherited.
    """

    def __init__(self, *args, index=None, columns=None, dtype=None, copy: bool = False):
        """
        Input data or address string, and convert it to OneData format.
        Now the input args supported:
            1. str: a string will be recognized as the location of a file, and input with the format correctly
            2. DataFrame & OneData
            3. Dictionary
            4. Numpy Array
            ...
        OneData inherits all the features of DataFrame.
        """

        data = {}
        if not args:
            pass
        elif isinstance(args[0], str):
            file_form = args[0].split('.')[-1]
            if file_form == 'csv':
                data = pd.read_csv(args[0])
            elif file_form == 'xls' or file_form == 'xlsx':
                data = pd.read_excel(args[0])
            elif file_form == 'json':
                data = pd.read_json(args[0])
            elif file_form == 'pkl':
                data = pd.read_pickle(args[0])
            elif file_form == 'hdf':
                data = pd.read_hdf(args[0])
            else:
                self._raise_format_error()
        else:
            data = args[0]
        try:
            super().__init__(data=data, index=index, columns=columns, dtype=dtype, copy=copy)
        except (ValueError, TypeError):
            self._raise_format_error()

    def __getitem__(self, key):
        """
        The rebuilt method of __getitem__, which will redirect the DataFrame and Series to OneData and OneSeries.
        """

        key = lib.item_from_zerodim(key)
        key = com.apply_if_callable(key, self)

        if is_hashable(key):
            if self.columns.is_unique and key in self.columns:
                if self.columns.nlevels > 1:
                    return OneSeries(self._getitem_multilevel(key))
                return OneSeries(self._get_item_cache(key))

        indexer = convert_to_index_sliceable(self, key)
        if indexer is not None:
            return OneData(self._slice(indexer, axis=0))

        if isinstance(key, DataFrame) or isinstance(key, OneData):
            return self.where(key)

        if com.is_bool_indexer(key):
            return OneData(self._getitem_bool_array(key))

        is_single_key = isinstance(key, tuple) or not is_list_like(key)

        if is_single_key:
            if self.columns.nlevels > 1:
                return self._getitem_multilevel(key)
            indexer = self.columns.get_loc(key)
            if is_integer(indexer):
                indexer = [indexer]
        else:
            if is_iterator(key):
                key = list(key)
            indexer = self.loc._get_listlike_indexer(key, axis=1, raise_missing=True)[1]

        if getattr(indexer, "dtype", None) == bool:
            indexer = np.where(indexer)[0]

        data = self._take_with_is_copy(indexer, axis=1)

        if is_single_key:
            if data.shape[1] == 1 and not isinstance(self.columns, ABCMultiIndex):
                data = data[key]

        return OneData(data)

    def show(self, info: bool = False,
             all_columns: bool = False,
             all_rows: bool = False,
             max_columns: int = None,
             max_rows: int = None,
             precision: int = None):
        """
        Display the info and details of data.
        The display options would be reset to the previous state after printing data, in other words, the
        options that was edited in show() would not be inherited.

        :param info: stay True if a display of information is required
        :param all_columns: reset the display of columns if columns equals to True
        :param all_rows: reset the display of rows if rows equals to True
        :param max_columns: the max columns of display
        :param max_rows: the max rows of display
        :param precision: a fast adjustment to the precision of display
        :return None
        """

        if info:
            self.info(memory_usage=False)
            print('\n\n - Shape: {}\n - Index: {}\n - Memory usage: {:.3f} MB\n'
                  .format(self.shape, ", ".join(self.columns), self.memory_usage().sum() / 1024 ** 2))
        original_max_columns = pd.get_option('display.max_columns')
        original_max_rows = pd.get_option('display.max_rows')
        original_display_precision = pd.get_option('display.precision')
        if max_columns:
            pd.set_option('display.max_columns', max_columns)
        if max_rows:
            pd.set_option('display.max_rows', max_rows)
        if all_columns:
            pd.set_option('display.max_columns', None)
        if all_rows:
            pd.set_option('display.max_rows', None)
        if precision:
            pd.set_option('display.precision', precision)
        print(self)
        pd.set_option('display.max_rows', original_max_rows)
        pd.set_option('display.max_columns', original_max_columns)
        pd.set_option('display.precision', original_display_precision)

    def make_dataset(self, train_frac: float = None,
                     random: bool = False,
                     random_seed: int = None):
        """
        Create the separated datasets based on the original one.
        The proportion of the train data and hold-out data should be specified.

        :param train_frac: the proportion of train data
        :param random: set True if random dataset is needed
        :param random_seed: the random seed of making dataset procession
        :returns train_data, test_data
        """

        if random:
            train_data = self.sample(frac=train_frac, random_state=random_seed)
            test_data = self.drop(train_data.index)
            return OneData(train_data), OneData(test_data)
        elif train_frac:
            index_num = int(self.shape[0] * train_frac)
            train_data = self.iloc[:index_num, :]
            test_data = self.iloc[index_num:, :]
            return OneData(train_data), OneData(test_data)
        else:
            self._raise_parameter_error(["train_frac"])

    def reverse(self, axis: str = 'row', reset_index: bool = False):
        """
        Method for reversing the dataset.

        :param axis: set 'row' if the order of the rows is to be reversed, and set 'column' if the order of the column
        is to be reversed
        :param reset_index: true for reset the index of data frame
        :param inplace: inplace modification if it sets True
        :return OneData
        """

        # the acceptable attributes
        row_attr = {"row", "r"}
        column_attr = {"column", "col", "c"}

        # match attributes
        if axis in row_attr:
            if reset_index:
                return OneData(self.loc[::-1].reset_index(drop=True))
            else:
                return OneData(self.loc[::-1])
        elif axis in column_attr:
            return OneData(self.loc[:, ::-1])
        else:
            self._wrong_parameter_error("axis")

    def summary(self, info: bool = True):
        """
        Return a summary of the whole dataset. And the stats from scipy is used to calculate the Entropy.

        :param info: stay True if a display of information is required
        :return the detail of summary
        """

        # summarize the data and build summary_info sheet
        summary_info = OneData(self.dtypes, columns=['dtypes'])
        summary_info = summary_info.reset_index()
        summary_info['Name'] = summary_info['index']
        summary_info = summary_info[['Name', 'dtypes']]
        summary_info['Missing'] = self.isnull().sum().values
        summary_info['Uniques'] = self.nunique().values
        summary_info['First Value'] = self.loc[self.index[0]].values
        summary_info['Second Value'] = self.loc[self.index[1]].values
        summary_info['Third Value'] = self.loc[self.index[2]].values

        # Add Entropy if scipy is installed
        stats = import_optional_dependency("scipy.stats")

        if stats:
            for name in summary_info['Name'].value_counts().index:
                summary_info.loc[summary_info['Name'] == name, 'Entropy'] = round(
                    stats.entropy(self[name].value_counts(normalize=True), base=2), 2)

        if info:
            pd.set_option('display.max_columns', None)
            print(f"Dataset Shape: {self.shape}")
            print(summary_info)
        return OneData(summary_info)

    def fill_na(
            self,
            value=None,
            method=None,
            axis=None,
            inplace=False,
            limit=None,
            downcast=None,
    ):
        """
        Fill the NaN values, which is an extended function of original fillna() and has more methods to fill
        the NaN data.

        :param method: the way to fill the NaN values, which contains mode and nan methods
        :param inplace: inplace modification
        :param value: inherit
        :param axis: inherit
        :param limit: inherit
        :param downcast: inherit
        """

        if method == 'mode':
            if inplace:
                for key, value in self.isnull().sum().items():
                    if value:
                        self[key].fillna(self[key].mode()[0], inplace=True)
            else:
                data = self
                for key, value in data.isnull().sum().items():
                    if value:
                        data[key].fillna(self[key].mode()[0], inplace=True)
                return data
        elif method == 'nan':
            if inplace:
                self.fillna(np.nan, inplace=True)
            else:
                data = self
                data = data.fillna(np.nan)
                return data
        else:
            return self.fillna(
                value=value,
                method=method,
                axis=axis,
                inplace=inplace,
                limit=limit,
                downcast=downcast,
            )

    def remove(self, column: list = None, row: list = None, inplace: bool = False):
        """
        The advanced function of DataFrame.drop, you can input a list of index to drop them.

        :param column: the list of column you want to remove
        :param row: the list of row you want to remove
        :param inplace: inplace remove or not
        """

        # inplace measurement
        if inplace:
            if column:
                self.drop(column, axis=1, inplace=True)
            if row:
                for n in row:
                    self.drop(n, inplace=True)
        # non inplace measurement
        else:
            data = self
            if column:
                data = data.drop(column, axis=1)
            if row:
                for n in row:
                    data = data.drop(n)
            return OneData(data)

    def map_values(self, exist, new, mapper=None):
        """
        Generate a new variable based on the calculating of exist variable using map().

        :param exist: the exist column of data
        :param new: the name of new column
        :param mapper: the mapper applied to the exist variable

        For example:
        >>       0    1    2
            0  1.0  2.0  3.0
            1  1.0  2.0  2.0
            2  1.0  2.0  2.0
            3  8.0  8.0  2.0

        >> data.add_var(0, 3, lambda x: x * 2)

        >>       0    1    2     3
            0  1.0  2.0  3.0   2.0
            1  1.0  2.0  2.0   2.0
            2  1.0  2.0  2.0   2.0
            3  8.0  8.0  2.0  16.0
        """

        data = self
        if mapper:
            data[new] = list(map(mapper, data[exist]))
        else:
            data[new] = data[exist]
        return data

    def mean(self, axis=None, skipna=None, level=None, numeric_only=None, **kwargs):
        """
        The mean method.

        :param axis: inherit
        :param skipna: inherit
        :param level: inherit
        :param numeric_only: inherit
        :param kwargs: inherit
        :return: OneSeries
        """

        return OneSeries(NDFrame.mean(self, axis, skipna, level, numeric_only, **kwargs))

    def reduce_mem_usage(self, use_float16: bool = False, info: bool = True):
        """
        Automatically distinguish the type of one single data and reset a suitable type.

        :param use_float16: use float16 or not
        :param info: stay True if a display of information is required
        """

        # the original memory usage
        start_mem = self.memory_usage().sum() / 1024 ** 2

        # reduce the memory usage
        for col in self.columns:
            if is_datetime(self[col]) or is_categorical_dtype(self[col]):
                continue
            col_type = self[col].dtype

            if col_type != object:
                c_min = self[col].min()
                c_max = self[col].max()
                if str(col_type)[:3] == "int":
                    if c_min > np.iinfo(np.int8).min and c_max < np.iinfo(np.int8).max:
                        self[col] = self[col].astype(np.int8)
                    elif c_min > np.iinfo(np.int16).min and c_max < np.iinfo(np.int16).max:
                        self[col] = self[col].astype(np.int16)
                    elif c_min > np.iinfo(np.int32).min and c_max < np.iinfo(np.int32).max:
                        self[col] = self[col].astype(np.int32)
                    elif c_min > np.iinfo(np.int64).min and c_max < np.iinfo(np.int64).max:
                        self[col] = self[col].astype(np.int64)
                else:
                    if use_float16 and c_min > np.finfo(np.float16).min and c_max < np.finfo(np.float16).max:
                        self[col] = self[col].astype(np.float16)
                    elif c_min > np.finfo(np.float32).min and c_max < np.finfo(np.float32).max:
                        self[col] = self[col].astype(np.float32)
                    else:
                        self[col] = self[col].astype(np.float64)
            else:
                self[col] = self[col].astype("category")

        if info:
            end_mem = self.memory_usage().sum() / 1024 ** 2
            print("Memory usage before optimization:\t{:.3f} MB".format(start_mem))
            print("Memory usage after optimization:\t{:.3f} MB".format(end_mem))
            print("Decreased by {:.1f}%".format(100 * (start_mem - end_mem) / start_mem))

    def iter(self, index=True, name="OneData"):
        """
        Iteration methods for fast using, the override of itertuples().

        :param index: If True, return the index as the first element of the tuple.
        :param name: The name of the returned named tuples or None to return regular.
        """

        arrays = []
        fields = list(self.columns)
        if index:
            arrays.append(self.index)
            fields.insert(0, "Index")
        arrays.extend(self.iloc[:, k] for k in range(len(self.columns)))

        if name is not None:
            itertuple = collections.namedtuple(name, fields, rename=True)
            return map(itertuple._make, zip(*arrays))

        return zip(*arrays)

    def r_append(self, other):
        """
        Append a OneData(DataFrame) or OneSeries(Series) to the right of dataset.
        WARNING: this is not a inplace method.

        :param other: the data
        :return: OneData
        """

        return OneData(pd.concat([self, other], axis=1))

    def l_append(self, other):
        """
        Append a OneData(DataFrame) or OneSeries(Series) to the left of dataset.
        WARNING: this is not a inplace method.

        :param other: the data
        :return: OneData
        """

        return OneData(pd.concat([other, self], axis=1))

    # ================== Plot Functions ================== #

    def line_plot(self, y,
                  x: str = None,
                  inherit: plt = None,
                  figsize: list = None,
                  title: str = None,
                  xlabel: str = None,
                  ylabel: str = None,
                  smooth: bool = False,
                  kind: str = 'cubic',
                  interval: int = 50,
                  legend_loc: str = 'upper left',
                  show: bool = True):
        """
        It's a fast plot function to generate graph rapidly.

        :param x: the x
        :param y: the y, which should be imported as list or str
        :param inherit: the inherit plt
        :param figsize: The size of figure.
        :param title: the title of plot
        :param xlabel: label of x
        :param ylabel: label of y
        :param smooth: Set it True if the curve smoothing needed.
        :param kind: The method for smooth.
        :param interval: define the number for smooth function.
        :param legend_loc: The location of the legend in plot.
        :param show: plt.show will run if true
        :return: plt
        """

        # Check if y is list or str
        if not isinstance(y, list) and not isinstance(y, str):
            self._raise_plot_format_error(["y"], "list or str")

        # previous configuration of plot
        plt = self._plot_prev_config(inherit, figsize, "seaborn-darkgrid")

        # data x pre-configuration process
        # x will be used if x is specified, otherwise the default index [0, len()] would be used
        if x:
            x = self[x]
        else:
            x = list(self.index)

        # data y pre-configuration process
        # y should import as a list for multiple series plot
        if isinstance(y, list):
            for name in y:
                plt = self._meta_line_plot(plt, self[name], x, smooth, kind, interval, name)
        else:
            plt = self._meta_line_plot(plt, self[y], x, smooth, kind, interval, y)

        # return for advanced adjustment
        return self._plot_post_config(plt, legend_loc, title, xlabel, ylabel, show)

    def count_plot(self, variable,
                   hue: str = None,
                   inherit: plt = None,
                   figsize: list = None,
                   title: str = None,
                   xlabel: str = None,
                   ylabel: str = None,
                   show: bool = True):
        """
        Generate the count graph.

        :param variable: The variable that should be counted
        :param hue: the hue parameter for using
        :param inherit: the inherit plt
        :param figsize: The size of figure.
        :param title: the title of plot
        :param xlabel: label of x
        :param ylabel: label of y
        :param show: plt.show will run if true
        :return: plt
        """

        # Check if y is list or str
        if not isinstance(variable, str):
            self._raise_plot_format_error(["variable"], "str")

        # previous configuration of plot
        plt = self._plot_prev_config(inherit, figsize, "seaborn-darkgrid")

        if hue:
            # unique value of hue
            unique = self[hue].unique().tolist()

            # generate the plot
            fig, axes = plt.subplots(1, len(unique))
            fig.set_size_inches(figsize)
            for index, ax in enumerate(axes):
                ax.set_xlabel(variable)
                ax.yaxis.set_major_formatter(mtick.FormatStrFormatter('%.2f'))
                ax.set_title(f"{ hue } = { unique[index] }")
                temp_data = self[self[hue] == unique[index]][variable].value_counts()
                plt.bar(list(temp_data.keys()), list(temp_data))
        else:
            count_base_data = self[variable].value_counts()
            plt.bar(list(count_base_data.keys()), list(count_base_data))

        # return for advanced adjustment
        return self._plot_post_config(plt, '', title, xlabel, ylabel, show)

    def corr_plot(self, parameters: list = None,
                  inherit: plt = None,
                  figsize: list = None,
                  title: str = None,
                  annot: bool = True,
                  show: bool = True):
        """
        Generate the correction graph.

        :param parameters: The parameters selected.
        :param inherit: the inherit plt
        :param figsize: The size of figure.
        :param title: the title of plot
        :param annot: Display the annotation or not.
        :param show: plt.show will run if true
        :return: plt
        """

        # Check if y is list or str
        if not isinstance(parameters, list):
            self._raise_plot_format_error(["parameters"], "list or None")

        # previous configuration of plot
        plt = self._plot_prev_config(inherit, figsize, "seaborn-dark")
        fig, ax = plt.subplots()

        data = self[parameters].corr()
        score = data.values
        col = data.columns
        length = len(col)
        im = ax.imshow(score, cmap='rocket_r')
        ax.xaxis.set_ticks_position('top')
        ax.set_xticks(np.arange(length))
        ax.set_yticks(np.arange(length))
        ax.set_xticklabels(col)
        ax.set_yticklabels(col)
        fig.colorbar(im, pad=0.03)

        # the annotation part
        if annot:
            for i in range(length):
                for j in range(length):
                    if score[i, j] > 0.4:
                        color = "w"
                    else:
                        color = "black"
                    ax.text(j, i, round(score[i, j], 2),
                            ha="center", va="center", color=color)

        # return for advanced adjustment
        return self._plot_post_config(plt, '', title, '', '', show)

    def hist_plot(self, variable: str = None,
                  hue: str = None,
                  inherit: plt = None,
                  figsize: list = None,
                  show: bool = True):
        """
        Generate the hist graph.

        :param variable: the variable to use
        :param hue: the second variable to use
        :param inherit: the inherit
        :param figsize: the size of figuration.
        :param show: show the graph if it's True
        :return: plt
        """

        # check the format of variable1 and variable2
        if not variable and not hue:
            self._raise_plot_value_error(["variable", "hue"])
        elif self[variable].dtype == 'object' or (hue and self[hue].dtype == 'object'):
            self._raise_plot_format_error(["variable", "hue"], "int or float")

        # previous configuration of plot
        plt = self._plot_prev_config(inherit, figsize, "seaborn-darkgrid")

        if hue:
            # unique value of variable2
            unique = self[hue].unique().tolist()

            # generate the plot
            fig, axes = plt.subplots(1, len(unique))
            fig.set_size_inches(figsize)
            for index, ax in enumerate(axes):
                ax.set_xlabel(variable)
                ax.yaxis.set_major_formatter(mtick.FormatStrFormatter('%.2f'))
                ax.set_title(f"{ hue } = { unique[index] }")
                temp_data = self[self[hue] == unique[index]][variable]
                ax.hist(temp_data, density=True, stacked=True)

        else:
            plt.hist(self[variable])

        # return for advanced adjustment
        return self._plot_post_config(plt, '', '', '', '', show)



"""
OneSeries is an extended variant of pandas.Seres, which also inherits all the pandas.Series
features and ready to use. It contains many useful methods for a better experience on data analysis.

WARNING: Because this module is still pre-alpha, so many features are unstable.
"""

import pandas as pd
from pandas import Series
from pandas import DataFrame
import matplotlib.pyplot as plt

# oneline module
from .plot import Plot
from ..tools.compat import import_optional_dependency


class OneSeries(Series, Plot):

    def __init__(self, data):
        """
        Inherit from pandas.Series.

        :param data: the import data
        """

        super().__init__(data=data)

    @property
    def _one_data(self):
        from .onedata import OneData
        return OneData

    def r_append(self, other):
        """
        Append another OneSeries data to the right of this series and return a new OneData.

        :param other: the other OneSeries
        :return: OneData
        """

        if isinstance(other, OneSeries):
            return self._one_data({self.name: self, other.name: other})
        elif isinstance(other, self._one_data) or isinstance(other, DataFrame):
            return self._one_data(pd.concat([self, other], axis=1))

    def l_append(self, other):
        """
        Append another OneSeries data to the left of this series and return a new OneData.

        :param other: the other OneSeries
        :return: OneData
        """

        if isinstance(other, OneSeries):
            return self._one_data({other.name: other, self.name: self})
        elif isinstance(other, self._one_data) or isinstance(other, DataFrame):
            return self._one_data(pd.concat([other, self], axis=1))

    def summary(self, info: bool = True):
        """
        Return a summary of the whole OneSeries dataset.
        the stats from scipy is used to calculate the Entropy.

        :param info: stay True if a display of information is required
        """

        # Add Entropy if scipy is installed
        stats = import_optional_dependency("scipy.stats")

        sum_info = {}
        length = self.shape[0]
        if str(self.dtype)[:3] == "int":
            sum_info['Unique'] = len(self.unique())
            sum_info['Unique(%)'] = sum_info['Unique'] / length
            sum_info['Missing'] = self.isnull().sum()
            sum_info['Missing(%)'] = sum_info['Missing'] / length
            sum_info['Means'] = self.sum() / length
            sum_info['Minimum'] = self.min()
            sum_info['Maximum'] = self.max()
            sum_info['Zeros'] = (self == 0).sum()
            sum_info['Zeros(%)'] = sum_info['Zeros'] / length
            if stats:
                sum_info['Entropy'] = round(
                    stats.entropy(self.value_counts(normalize=True), base=2), 2)
            sum_info['Memory Size(KB)'] = self.memory_usage() / 1024

            if info:
                print('Unique: {}({:.2f}%)'.format(sum_info['Unique'], sum_info['Unique(%)'] * 100))
                print('Missing: {}({:.2f}%)'.format(sum_info['Missing'], sum_info['Missing(%)'] * 100))
                print('Zeros: {}({:.2f}%)'.format(sum_info['Zeros'], sum_info['Zeros(%)'] * 100))
                print('Means: {:.2f}'.format(sum_info['Means']))
                print('Minimum: {}'.format(sum_info['Minimum']))
                print('Maximum: {}'.format(sum_info['Maximum']))
                if stats:
                    print('Entropy: {}'.format(sum_info['Entropy']))
                print('Memory Size: {:.1f}KB'.format(sum_info['Memory Size(KB)']))

        elif str(self.dtype) == "object":
            sum_info['Unique'] = len(self.unique())
            sum_info['Unique(%)'] = sum_info['Unique'] / length
            sum_info['Missing'] = self.isnull().sum()
            sum_info['Missing(%)'] = sum_info['Missing'] / length
            if stats:
                sum_info['Entropy'] = round(
                    stats.entropy(self.value_counts(normalize=True), base=2), 2)
            sum_info['Memory Size(KB)'] = self.memory_usage() / 1024

            if info:
                print('Unique: {}({:.2f}%)'.format(sum_info['Unique'], sum_info['Unique(%)'] * 100))
                print('Missing: {}({:.2f}%)'.format(sum_info['Missing'], sum_info['Missing(%)'] * 100))
                if stats:
                    print('Entropy: {}'.format(sum_info['Entropy']))
                print('Memory Size: {:.2f}KB'.format(sum_info['Memory Size(KB)']))

        return self._one_data(sum_info, index=[0])

    def reverse(self, reset_index: bool = False):
        """
        Method for reversing the dataset.

        :param reset_index: true for reset the index of series
        """

        if reset_index:
            return self.loc[::-1].reset_index(drop=True)
        else:
            return self.loc[::-1]

    def shuffle(self, reset_index: bool = False, random_seed: int = None):
        """
        A non-inplace shuffle method.

        :param reset_index: reset the index if it sets True
        :param random_seed: the random seed of shuffle
        :return: OneSeries
        """

        if reset_index:
            return self.sample(frac=1, random_state=random_seed).reset_index(drop=True)
        else:
            return self.sample(frac=1, random_state=random_seed)

    def top(self, n: int = 1):
        """
        Return the top of the values.

        :param n: the number of value to return
        """

        return self.value_counts().nlargest(n=n)

    def bottom(self, n: int = 1):
        """
        Return the bottom of the values.

        :param n: the number of value to return
        """

        return self.value_counts().nsmallest(n=n)

    def to_frame(self, name=None):
        """
        Convert Series to OneData.

        :param name: the name of series
        :return: oneline.OneData
        """

        if name is None:
            return self._onedata(self)
        else:
            return self._onedata({name: self})

    # ========================= Plot ========================= #

    def count_plot(self,
                   inherit: plt = None,
                   figsize: list = None,
                   title: str = None,
                   xlabel: str = None,
                   ylabel: str = None,
                   show: bool = True):
        """
        Generate the count graph based on the data.

        :param inherit: the inherit plt
        :param figsize: The size of figure.
        :param title: the title of plot
        :param xlabel: label of x
        :param ylabel: label of y
        :param show: plt.show will run if true
        """

        # previous configuration of plot
        plt = self._plot_prev_config(inherit, figsize, "seaborn-darkgrid")

        count_base_data = self.value_counts()
        plt.bar(list(count_base_data.keys()), list(count_base_data))

        # return for advanced adjustment
        return self._plot_post_config(plt, '', title, xlabel, ylabel, show)

    def line_plot(self,
                  inherit: plt = None,
                  figsize: list = None,
                  title: str = None,
                  xlabel: str = None,
                  ylabel: str = None,
                  smooth: bool = False,
                  kind: str = 'cubic',
                  interval: int = 50,
                  legend_loc: str = 'upper left',
                  show: bool = True):
        """
        It's a fast plot function to generate line_plot graph rapidly.

        :param inherit: the inherit plt
        :param figsize: The size of figure.
        :param title: the title of plot
        :param xlabel: label of x
        :param ylabel: label of y
        :param smooth: Set it True if the curve smoothing needed.
        :param kind: The method for smooth.
        :param interval: define the number for smooth function.
        :param legend_loc: The location of the legend in plot.
        :param show: plt.show will run if true
        """

        # previous configuration of plot
        plt = self._plot_prev_config(inherit, figsize, "seaborn-darkgrid")

        # data x pre-configuration process
        x = list(self.index)

        # data y pre-configuration process
        plt = self._meta_line_plot(plt, self, x, smooth, kind, interval, None)

        # return for advanced adjustment
        return self._plot_post_config(plt, legend_loc, title, xlabel, ylabel, show)
